#/* Copyright 2008-2014 Research Foundation State University of New York   */

#/* This file is part of QUB Express.                                      */

#/* QUB Express is free software; you can redistribute it and/or modify    */
#/* it under the terms of the GNU General Public License as published by   */
#/* the Free Software Foundation, either version 3 of the License, or      */
#/* (at your option) any later version.                                    */

#/* QUB Express is distributed in the hope that it will be useful,         */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
#/* GNU General Public License for more details.                           */

#/* You should have received a copy of the GNU General Public License,     */
#/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
#/* <http://www.gnu.org/licenses/>.                                        */

from __future__ import with_statement

from cStringIO import StringIO
from qubx.hist import *
from qubx.model import *
from qubx_single.rates import *
import qubx.fit
import qubx.pyenv
import qubx.pyenvGTK
import qubx.table
import qubx.toolspace
import numpy
import qubx.fast.model
from qubx.fast.fast_utils import c_double, c_double_p, cdata, buffer_as_array_of_int
from qubx.settings import Property, Propertied

NBIN_DUR = 50
LBL_TDEAD = 'Tdead (ms)'

MAX_CHANNEL_COUNT = 8
MAX_DURHIST_CLASSES = 32

COLOR_TCRIT = ('qubx_single.dur_hist.tcrit', (1, .2, .2, .7))
ColorInfo[COLOR_TCRIT[0]].label = 'Figures:DurHist Tcrit line'


@Propertied(Property('bins', 96, 'number of bins'),
            Property('smooth', True, 'True to spread short events among adjacent bins, adjusting for sampling error'),
            Property('sqrt', True, 'True to display sqrt(y axis)'))
class DurHistFace(HistFace):
    __explore_featured = ['lo', 'hi', 'auto', 'nbTimeConst', 'signals', 'controls', 'save_as',
                          'copy_to_clipboard', 'copy_image_to_clipboard', 'save_to', 'fit_all_exponentials']
    def __init__(self):
        HistFace.__init__(self, 'DurHist', 'QubX.Figures.DurHist')
        self.__ref = Reffer()
        self.propertied_connect_settings('qubx_single.dur_hist')
        self.__lo = 0.0
        self.__hi = 1.0
        self.__auto = True
        self.__in_all_exp = False
        self.__last_path = ''
        self.file_i_bins = 0
        self.file_i_bars = 1
        self.file_right_edge = True
        self.file_sqrt = True
        self.file_norm = False
        self.cancel_all_exp = False
        self.OnChangePage += self.__ref(self.__onChangePage)
        self.OnAddPlot += self.__ref(self.__onAddPlot)
        self.OnRemovingPlot += self.__ref(self.__onAddPlot)
        self.nbTimeConst = qubx.notebook.NbItems("All time constants", self.global_name and '%s.nbTimeConst'%self.global_name or '')
        self.signals = SinglePlex(SingleSignals(self.QubX.DataSource, self.QubX.Data, self.QubX.Modeling.Stimulus))
        self.signals.OnInvalidate += self.__ref(self.__on_invalidata)
        self.QubX.Modeling.MILRates.model_prep.OnInvalidate += self.__ref(self.__on_invalimodel)
        self.QubX.Modeling.MILRates.rates.OnInvalidate += self.__ref(self.__on_invalirates)

        self.controls = qubx.faces.Face("DurHist", global_name='QubX.About.DurHist')
        self.controls.show()
        line = gtk.HBox(True)
        line.show()
        lbl = gtk.Label('Bins:')
        lbl.show()
        line.pack_start(lbl)
        self.txtBins = qubx.GTK.NumEntry(self.bins, acceptIntGreaterThan(1))
        self.propertied_connect_NumEntry('bins', self.txtBins)
        self.txtBins.show()
        line.pack_start(self.txtBins)
        self.controls.pack_start(line, False, True)
        line = gtk.HBox(True)
        line.show()
        line = gtk.HBox()
        line.show()
        self.chkAuto = gtk.CheckButton('auto')
        self.chkAuto.set_active(self.__auto)
        self.chkAuto.connect('toggled', self.__onToggleAuto)
        self.chkAuto.show()
        line.pack_start(self.chkAuto, False, True)
        lbl = gtk.Label('lo:')
        lbl.show()
        line.pack_start(lbl, False, True)
        self.txtLo = qubx.GTK.NumEntry(self.__lo, format='%.2g')
        self.txtLo.set_width_chars(4)
        self.txtLo.OnChange += self.__ref(self.__onChangeLo)
        self.txtLo.show()
        line.pack_start(self.txtLo, False, True)
        lbl = gtk.Label('hi:')
        lbl.show()
        line.pack_start(lbl, False, True)
        self.txtHi = qubx.GTK.NumEntry(self.__hi, format='%.2g')
        self.txtHi.set_width_chars(4)
        self.txtHi.OnChange += self.__ref(self.__onChangeHi)
        self.txtHi.show()
        line.pack_start(self.txtHi, False, True)
        self.controls.pack_start(line, False, True)
        self.chkSmooth = gtk.CheckButton('Smooth')
        self.propertied_connect_check('smooth', self.chkSmooth)
        self.chkSmooth.show()
        self.controls.pack_start(self.chkSmooth, False, True)
        self.chkSqrt = gtk.CheckButton('Sqrt ordinate')
        self.propertied_connect_check('sqrt', self.chkSqrt)
        self.chkSqrt.show()
        self.controls.pack_start(self.chkSqrt, False, True)
        line = gtk.HBox()
        line.show()
        self.btnSave = gtk.Button('Save...')
        self.btnSave.connect('clicked', self.__onClickSave)
        self.btnSave.show()
        line.pack_start(self.btnSave, False, True)
        self.btnCopy = gtk.Button('Copy')
        self.btnCopy.connect('clicked', self.__onClickCopy)
        self.btnCopy.show()
        line.pack_start(self.btnCopy, False, True)
        self.btnCopyImage = gtk.Button('Image...')
        self.btnCopyImage.connect('clicked', self.__onClickCopyImage)
        self.btnCopyImage.show()
        line.pack_start(self.btnCopyImage, False, True)
        self.controls.pack_start(line, False, True)
        scr = gtk.ScrolledWindow()
        scr.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scr.show()
        self.txtStim = gtk.TextView()
        self.outStim = qubx.GTK.TextViewAppender(self.txtStim)
        self.txtStim.show()
        scr.add(self.txtStim)
        self.controls.pack_end(scr, True, True)

        self.__need_data = True
        self.__need_model = True
        self.__need_rates = True
        self.robot.model = QubModel()
        self.robot.OnInterrupt += self.__ref(self.__onInterrupt)

        # generally short-running, so it takes more time to create and remove the progress bar:
        self.robot.on_start_item = lambda: None
        self.robot.on_finish_item = lambda: None

        gobject.idle_add(self.update)
        
        self.Ns = self.Nc = 0
        self.path = None
    def propertied_set(self, value, name):
        super(DurHistFace, self).propertied_set(value, name)
        if name in ('bins', 'smooth', 'sqrt'):
            self.__on_invalidata()
    def set_auto(self, x):
        redo = x != self.__auto
        self.__auto = x
        self.chkAuto.set_active(x)
        if redo:
            self.__need_data = True
            self.update()
    auto = property(lambda self: self.__auto, lambda self, x: self.set_auto(x))
    def set_lo(self, x, auto=False):
        self.__lo = x
        self.txtLo.setValue(x)
        if not auto:
            self.auto = False
            self.__need_data = True
            self.update()
    lo = property(lambda self: self.__lo, lambda self, x: self.set_lo(x))
    def set_hi(self, x, auto=False):
        self.__hi = x
        self.txtHi.setValue(x)
        if not auto:
            self.auto = False
            self.__need_data = True
            self.update()
    hi = property(lambda self: self.__hi, lambda self, x: self.set_hi(x))
    def __onToggleAuto(self, chk):
        qubx.pyenv.env.OnScriptable('%s.auto = %s' % (self.global_name, repr(chk.get_active())))
        self.auto = chk.get_active()
    def __onChangeLo(self, txt, val):
        qubx.pyenv.env.OnScriptable('%s.lo = %s' % (self.global_name, repr(val)))
        self.lo = val
    def __onChangeHi(self, txt, val):
        qubx.pyenv.env.OnScriptable('%s.hi = %s' % (self.global_name, repr(val)))
        self.hi = val
    def __onClickSave(self, btn):
        fname = qubx.GTK.SaveAs('Save as...', self.path or "", 'DurHist', self.save_as)
        if fname:
            self.path = os.path.split(fname)[0]
    def save_as(self, fname):
        qubx.pyenv.env.OnScriptable('%s.save_as(%s)' % (self.global_name, repr(fname)))
        self.save_to(open(fname, 'w'))
    def __onClickCopy(self, btn):
        qubx.pyenv.env.OnScriptable('%s.copy_to_clipboard()' % self.global_name)
        self.copy_to_clipboard()
    def copy_to_clipboard(self):
        buf = cStringIO.StringIO()
        self.save_to(buf)
        gtk.clipboard_get('CLIPBOARD').set_text(buf.getvalue())
    def __onClickCopyImage(self, btn):
        if gtk.RESPONSE_ACCEPT == TheDurHistCopyDialog.run(self.plots[0], len(self.plots)):
            w = TheDurHistCopyDialog.copy_width
            h = TheDurHistCopyDialog.copy_height
            c = TheDurHistCopyDialog.copy_class
            qubx.pyenv.env.OnScriptable('%s.copy_image_to_clipboard(w=%s, h=%s, c=%s)' %
                                        (self.global_name, repr(w), repr(h), repr(c)))
            self.copy_image_to_clipboard(w, h, c)
    def copy_image_to_clipboard(self, w, h, c):
        pixmap = gdk.Pixmap(self.window, w, h, -1)
        ctx = pixmap.cairo_create()
        ctx.set_source_rgb(1,1,1)
        ctx.paint()
        self.plots[c].draw_to_context(ctx, w, h)
        del ctx
        pixbuf = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, w, h)
        pixbuf.get_from_drawable(pixmap, gdk.colormap_get_system(), 0, 0, 0, 0, -1, -1)
        clipboard = gtk.clipboard_get('CLIPBOARD')
        clipboard.set_image(pixbuf)
        del pixmap
        del pixbuf
        qubx.pyenv.env.gc_collect_on_idle()
        self.plots[c].redraw_canvas() # re-calculate geometry for mouse
    def onShow(self, showing):
        HistFace.onShow(self, showing)
        if showing:
            self.QubX.About.append_face(self.controls)
            self.QubX.About.show_face('DurHist')
        else:
            self.QubX.About.remove_face( self.QubX.About.index('DurHist') )
    def qubx_finalize(self):
        self.robot.stop()
    def __onInterrupt(self, robot, cancel):
        if self.__in_all_exp:
            self.cancel_all_exp = True
            cancel()
    def __on_invalidata(self, *args):
        self.__need_data = True
        # self.robot.interrupt()
        self.update()
    def __on_invalimodel(self, *args):
        self.__need_model = True
        # self.robot.interrupt()
        self.update()
    def __on_invalirates(self, *args):
        self.__need_rates = True
        # self.robot.interrupt()
        self.update()
    def save_to(self, f, scls=None):
        sc = scls
        if sc == None:
            sc = self.page
        for c,plot in enumerate(self.plots):
            hist = self.histrec[sc][c]
            f.write('Binmax\tlog10(Binmax)\t')
            if self.sqrt:
                f.write('sqrt(count%i/total)\t'%(c+1))
            else:
                f.write('count%i/total\t'%(c+1))
            if hist.pdf != None:
                f.write('PDF%i\t'%(c+1))
            f.write('\t'*len(hist.components))
        f.write('\n')
        for b in xrange(self.bins):
            for c,plot in enumerate(self.plots):
                hist = self.histrec[sc][c]
                f.write('%.8g\t' % hist.bins[b])
                f.write('%.8g\t' % log10(hist.bins[b]))
                f.write('%.8g\t' % hist.bars[b])
                if hist.pdf != None:
                    f.write('%.8g\t' % hist.pdf[b])
                for m in hist.components:
                    f.write('%.8g\t' % m[b])
            f.write('\n')
    def __onChangePage(self, index):
        if (0 <= index < len(self.histrec)) and len(self.histrec[index]):
            self.txtStim.get_buffer().set_text('\n\n'.join([self.histrec[index][0].txt_stim] + [hh.txt_time_const for hh in self.histrec[index]]))
    def __onAddPlot(self, index):
        plot = self.plots[index]
        ref = plot.xref = Reffer()
        plot.OnDraw += ref(bind_with_args_before(self.__onDrawPlot, index))
        if index == 0:
            self.plots[0].subNotebook.items.append(self.nbTimeConst)
        plot.nbTimeConst = qubx.notebook.NbDynText('Time constants', '%s.plots[%i].nbTimeConst' % (self.global_name, index), bind(self.__nb_time_const, index))
        plot.subNotebook.items.append(plot.nbTimeConst)
        self.nbTimeConst.items.insert(index, plot.nbTimeConst)
        plot.layAllExponentials = qubx.toolspace.Layer(x=-20, y=-2.5*qubx.fit_space.LINE_EMS, w=19, h=qubx.fit_space.LINE_EMS, cBG=plot.controls.cLayer)
        plot.actAllExponentials = bind(self.__onClickAllExponentials, plot)
        plot.layAllExponentials.add_sublayer(qubx.toolspace.SubLayer_Label('Fit all exponentials...', 0, 1, x=0, y=0, w=15, h=qubx.fit_space.LINE_EMS,
                                                                                        color=plot.controls.cButton, action=plot.actAllExponentials))
        plot.layAllExponentials.add_sublayer(qubx.toolspace.SubLayer_PNG(os.path.join(qubx.global_namespace.app_path, 'icons', 'folder.png'), x=16, y=0, w=qubx.fit_space.LINE_EMS, h=qubx.fit_space.LINE_EMS, tooltip="Open histogram from a file for fitting...",
                                                                                      action=ref(bind(self.__open_file, index))))
        plot.layAllExponentials.add_sublayer(qubx.toolspace.SubLayer_Label('?', 0, 1, x=18, y=0, w=1, h=qubx.fit_space.LINE_EMS, tooltip="About this algorithm...",
                                                                                        color=plot.controls.cButton, action=ref(self.__show_all_exponentials_info)))
        plot.controls.add_layer(plot.layAllExponentials)
        plot.layAllExpoLLR = qubx.toolspace.Layer(x=-16, y=-8*qubx.fit_space.LINE_EMS, w=15, h=qubx.fit_space.LINE_EMS, cBG=plot.controls.cLayer)
        plot.layAllExpoLLR.add_sublayer(qubx.toolspace.SubLayer_Label('LLR: ', 1, 0, x=0, y=0, w=6, h=qubx.fit_space.LINE_EMS, color=plot.controls.cText))
        plot.subAllExpoLLR = qubx.toolspace.SubLayer_Label('', -1, 0, x=7, y=0, w=8, h=qubx.fit_space.LINE_EMS, color=plot.controls.cNumber)
        plot.layAllExpoLLR.add_sublayer(plot.subAllExpoLLR)
        plot.showingLLR = false
        plot.layFileInfo = qubx.toolspace.Layer(x=-25, y=2, w=19, h=qubx.fit_space.LINE_EMS, cBG=plot.controls.cLayer)
        plot.subFileName = qubx.toolspace.SubLayer_Label('', 0, 1, x=0, y=0, w=17, h=qubx.fit_space.LINE_EMS,
                                                                      color=plot.controls.cText,
                                                                      action=ref(bind(self.__open_file, index)))
        plot.layFileInfo.add_sublayer(plot.subFileName)
        plot.subFileClose = qubx.toolspace.SubLayer_Label('x', 0, 1, x=18, y=0, w=1, h=qubx.fit_space.LINE_EMS,
                                                                       color=plot.controls.cButton,
                                                                       action=ref(bind(self.__close_file, index)))
        plot.layFileInfo.add_sublayer(plot.subFileClose)
        plot.has_file = false
    def __onRemovingPlot(self, index):
        del self.nbTimeConst.items[index]
    def __nb_time_const(self, index):
        hist = self.histrec[self.page][index]
        return hist.txt_time_const
    def __onDrawPlot(self, context, w, h, index):
        hist = self.histrec[self.page][index]
        if not hist.tcrit:
            return
        plot = self.plots[index]
        if plot.layerset == plot.controls:
            return
        context.set_source_rgba(*plot.appearance.color(COLOR_TCRIT))
        context.set_line_width(2*plot.appearance.line_width)
        for t in hist.tcrit:
            x = plot.t2x(log10(1e3*t))
            context.move_to(x, 0)
            context.line_to(x, h)
            context.stroke()
    def robot_update(self):
        #qubx.pyenv.env.stderr.write("%s\n"%('robot begin'))
        self.robot.progress = 0
        self.robot.new_data = self.__need_data
        if self.__need_data:
            segs, mcls, scls, names, sampling_ms, td_ms = self.robot.gui_call_recv(self.signals.request, self.robot)
            shown_sc = max(0, min(int(self.page), len(scls)-1))
            self.robot_update_data(segs, mcls, scls, names, sampling_ms, td_ms, shown_sc)
        self.robot_update_pdfs()
    def robot_update_data(self, segs, mcls, scls, names, sampling_ms, td_ms, shown_sc):
        self.robot.names = names
        sampling = sampling_ms * 1e-3
        auto, lo, hi = self.auto, self.lo, self.hi
        lo -= td_ms
        hi -= td_ms
        lo *= 1e-3
        hi *= 1e-3
        if not segs: return
        self.robot.progress = 20
        per_scls = events_per_scls(segs, mcls, scls)
        if not per_scls: return
        self.robot.progress = 30
        if auto:
            t_min = [len(du) and min(du) or 0.0 for cl, du in per_scls]
            t_max = [len(du) and max(du) or 0.0 for cl, du in per_scls]
            if len(t_min):
                with self.robot.main_hold:
                    self.set_lo(t_min[shown_sc]*1e3+td_ms, auto=True)
                    self.set_hi(t_max[shown_sc]*1e3+td_ms, auto=True)
        else:
            t_min = len(scls) * [lo]
            t_max = len(scls) * [hi]
        self.robot.Nc = min(MAX_DURHIST_CLASSES, max(len(cl) and max(cl) or 0 for cl, du in per_scls) + 1)
        stimulus = [ dict([(name, scval) for name, scval in izip(names, scvals)]) for scvals in scls ]
        #Hist = self.smooth and DurHist_Smooth or DurHist
        Hist = self.smooth and DurHist_Smooth_lib or DurHist_lib
        cap_stim = []
        txt_stim = []
        for sc in xrange(len(per_scls)):
            buf_stim = StringIO()
            cap = ""
            for i, nm in enumerate(names):
                if i > 0:
                    buf_stim.write('%s: %.6g\n' % (nm, stimulus[sc][nm]))
                    if i > 1:
                        cap = cap + ', '
                    cap = cap + '%.3g' % stimulus[sc][nm]
            cap = cap and (' at '+cap)
            txt_stim.append(buf_stim.getvalue())
            cap_stim.append(cap)
        
        def make_rec(hist_out, sc, c):
            bins, bars = hist_out
            logbins = numpy.log10(bins)
            if self.sqrt:
                bars = numpy.sqrt(bars)
            caption = 'DurHist of class %d%s'%(c, cap_stim[sc])
            return HistRec(caption, caption, logbins, bars, linbins=bins,
                           nb_headers=['log10_ms', 'count_per_n'], nb_xlabel='log10(ms)', nb_ylabel='count / n',
                           data_colorref=qubx.modelGTK.COLOR_CLASS(c), txt_stim=txt_stim[sc], txt_time_const="")
        hist_sc_cl = []
        for sc, data in enumerate(per_scls):
            bins = DurHist_Bins([data], td_ms, t_min[sc], t_max[sc], self.bins)
            hist_of_sc = []
            for c in xrange(self.robot.Nc):
                if (c < len(self.plots)) and self.plots[c].has_file:
                    hist_of_sc.append(self.histrec[sc][c])
                else:
                    hist_of_sc.append(make_rec(Hist(c, [data], td_ms, sampling, bins), sc, c))
            hist_sc_cl.append(hist_of_sc)

        self.robot.progress = 45
        with self.robot.main_hold:
            self.stimulus = stimulus
            self.sig_names = names
            self.histrec = hist_sc_cl
            self.Ns = len(scls)
            self.td = td_ms
            self.__need_data = False
        # windows can't handle off-thread window creation
        gobject.idle_add(self.set_page_count, self.Ns)
        gobject.idle_add(self.update_plots)
        
    def robot_update_pdfs(self):
        self.robot.progress = 50
        if self.robot.new_data or self.__need_model or self.__need_rates:
            if self.__need_model:
                self.robot.model.undoStack.clear()
                mt = self.QubX.Modeling.MILRates.model_prep.val.modeltree
                if mt.find('States').find('State').isNull:
                    return
                self.robot.model.from_tree( self.QubX.Modeling.MILRates.model_prep.val.modeltree )
            elif self.__need_rates:
                self.__need_rates = False
                rrates = self.robot.model.rates
                K0, K1, K2 = self.QubX.Modeling.MILRates.rates.val
                if not (K0 is None):
                    for i in xrange(rrates.size):
                        a, b = [rrates.get(i, field) for field in ['From', 'To']]
                        rrates.set(i, 'k0', K0[a,b])
                        rrates.set(i, 'k1', K1[a,b])
                        rrates.set(i, 'k2', K2[a,b])
            self.robot.progress = 70
            self.robot.model.channelCount = min(MAX_CHANNEL_COUNT, self.robot.model.channelCount)
            single_model = ModelToFast(self.robot.model, self.robot.names, printer)
            multi_model = qubx.fast.model.MultiModel(single_model)
            multi_model.update()
            mm = multi_model.multi_model
            clazz = buffer_as_array_of_int(mm[0].clazz, mm[0].Nstate)
            
            classmap = sorted(list(set(st.Class for st in self.robot.model.states)))
            
            stimclasses_ = numpy.zeros(shape=(self.Ns, len(self.robot.names)), dtype='float64')
            for s in xrange(self.Ns):
                for ni, nm in enumerate(self.robot.names):
                    stimclasses_[s,ni] = self.stimulus[s][nm]
            #print stimclasses_
            stimclasses = cdata(stimclasses_, c_double_p)
            ratesm = qubx.fast.model.StimRates(mm, self.Ns, len(self.robot.names), stimclasses, None, 0.0) # skip ratesm.AA generation
            ratesm.setup_QQ()
            for s in xrange(self.Ns):
                Q = numpy.matrix(ratesm.QQ[s])
                #print Q
                for c in xrange(self.plot_count):
                    if self.robot.model.channelCount == 1:
                        if c in classmap:
                            cl = classmap.index(c)
                        else:
                            continue
                    elif c in clazz:
                        cl = c
                    else:
                        continue
                    hist = self.histrec[s][c]
                    if (not (hist.bins is None)) and (len(hist.bins) > 1):
                        hist.pdf, hist.components, hist.time_const_tau, hist.time_const_amp, hist.tcrit = PDF(Q, clazz, cl, self.td, hist.linbins, not self.QubX.Modeling.MILRates.optimizing)
                        if self.sqrt:
                            hist.pdf = sqrt(hist.pdf)
                            hist.components = [sqrt(cc) for cc in hist.components]
                        hist.txt_time_const = "Class %i:\nTime Constants: (Tau [ms], Amp)\n" % c + \
                            "\n".join(["%.4g, %.4g" % (1e3*tc_tau, tc_amp) for (tc_tau, tc_amp) in izip(hist.time_const_tau, hist.time_const_amp)])
                        if hist.tcrit:
                            hist.txt_time_const += "\n\nTcrit:\n" + ", ".join("%.4g" % (1e3*crit) for crit in hist.tcrit) + "\n\n"
            #qubx.pyenv.env.stderr.write("%s\n"%('robot model'))
            single_model.dispose() # work around callback-related memory leak
            self.__need_model = self.__need_rates = False
        #qubx.pyenv.env.stderr.write("%s\n"%('robot done'))
        gobject.idle_add(self.update_plots)
    def __onClickAllExponentials(self, plot):
        if self.__in_all_exp:
            return
        qubx.pyenv.env.OnScriptable('QubX.Figures.DurHist.fit_all_exponentials(%i)' % self.plots.index(plot))
        self.fit_all_exponentials(self.plots.index(plot))
    def fit_all_exponentials(self, index):
        if self.__in_all_exp:
            return
        self.__in_all_exp = True
        self.cancel_all_exp = False
        plot = self.plots[index]
        plot.controls.remove_layer(plot.layAllExponentials)
        ### todo: prompt, remember properties:
        is_sqrt = plot.is_sqrt if plot.has_file else self.sqrt
        self.robot.do(self.robot_all_exponentials, plot, is_sqrt, plot.controls.xx, plot.controls.yy, ALL_EXPONENTIALS_SIGNIFICANCE)
    def robot_all_exponentials(self, plot, is_sqrt, xx, yy, significance): # ignoring weight expression
        pool = multiprocessing.Pool()
        def map_func(func, args):
            result = [x for x in pool.imap(func, args)]
            return result
        try:
            expo = AllExponentials()
            params = [1.0] * len(expo.params)
            if is_sqrt:
                yy = yy**2
            for pct in expo.run_generator(xx, yy, params, map_func=map_func, significance=significance):
                gobject.idle_add(self.__show_expo, plot, is_sqrt, params[:], pct > 99.5)
        except:
            traceback.print_exc()
        pool.close()
        gobject.idle_add(self.__done_expo, plot, expo.report_buf.getvalue(), expo.llr) ### plus e.g. N event required
    def __show_expo(self, plot, is_sqrt, params, can_fit):
        N = int(round(params[0]))
        robot = plot.controls.robot
        robot.set_curve(AllExponentials_Display, "%i %sExponential%s, see Landowne et al, 2013"%(N, is_sqrt and "(sqrt) " or "", (N>1) and "s" or ""))
        for i in xrange(N):
            robot.set_param("tau%i"%(i+1), params[2*i+1], 1e-16, UNSET_VALUE, can_fit)
            robot.set_param("area%i"%(i+1), params[2*i+2], 0.0, UNSET_VALUE, can_fit)
    def __done_expo(self, plot, report, llr):
        if not plot.showingLLR:
            plot.controls.add_layer(plot.layAllExpoLLR)
        plot.subAllExpoLLR.label = '%.2g' % llr
        # plot.controls.robot.fit()
        plot.controls.add_layer(plot.layAllExponentials)
        print report
        self.__in_all_exp = False
        pass ### show more info
    def __open_file(self, index):
        qubx.GTK.Open("Open duration histogram...", self.__last_path, bind_with_args_before(self.__do_open, index), parent=self.parent_window)
    def __do_open(self, path, index):
        plot = self.plots[index]
        self.__last_path, fname = os.path.split(path)
        headers, types, columns = qubx.table.read_table_text(open(path, 'r').read())
        i_bins, i_bars, right_edge, is_sqrt, is_norm = self.prompt_file_columns(headers, types, columns)
        if i_bins is None:
            return
        self.__close_file(index, False)
        plot.subFileName.label = fname if (len(fname) < 15) else (fname[:15]+'...')
        if not plot.has_file:
            plot.has_file = True
            plot.controls.add_layer(plot.layFileInfo)
        plot.is_sqrt = is_sqrt
        bins = columns[i_bins]
        bars = columns[i_bars]
        if is_norm:
            total = 0.0
            for x in bars:
                total += x
            if total:
                bars = [x/total for x in bars]
        if not right_edge:
            w = (bins[2] - bins[1])
            bins = [b+w for b in bins]
        linbins = numpy.power(10, bins)
        self.histrec[self.page][index] = HistRec(fname, fname, bins, bars, linbins=linbins)
        self.update_plots()
    def __close_file(self, index, stayingShut=True):
        plot = self.plots[index]
        if plot.has_file and stayingShut:
            plot.has_file = False
            plot.controls.remove_layer(plot.layFileInfo)
            self.__on_invalidata()
    def prompt_file_columns(self, headers, types, columns):
        dlg = gtk.Dialog('Import duration histogram...', self.parent_window, gtk.DIALOG_MODAL)
        h = pack_item(gtk.HBox(), dlg.vbox)
        pack_label('Bin column: ', h)
        print headers, types, len(columns)
        colChoices = ['%s: %s, ...' % (name, ', '.join(['%.3g'%v for v in columns[i][:3]])) for i, name in enumerate(headers)]
        print colChoices
        mnuBin = pack_item(qubx.GTK.StaticComboList(colChoices), h)
        mnuBin.active_i = self.file_i_bins
        chkRight = pack_check('contains right edge of each bin (otherwise, the left edge)', dlg.vbox, self.file_right_edge)
        h = pack_item(gtk.HBox(), dlg.vbox)
        pack_label('Bar column: ', h)
        mnuBar = pack_item(qubx.GTK.StaticComboList(colChoices), h)
        mnuBar.active_i = self.file_i_bars
        chkSqrt = pack_check('contains sqrt(count/total) (otherwise, count/total)', dlg.vbox, self.file_sqrt)
        chkNorm = pack_check('normalize (divide by sum)', dlg.vbox, self.file_norm)
        dlg.add_buttons('OK', gtk.RESPONSE_ACCEPT, 'Cancel', gtk.RESPONSE_REJECT)
        response = dlg.run()
        if response == gtk.RESPONSE_ACCEPT:
            self.file_i_bins = mnuBin.active_i
            self.file_i_bars = mnuBar.active_i
            self.file_right_edge = chkRight.get_active()
            self.file_sqrt = chkSqrt.get_active()
            self.file_norm = chkNorm.get_active()
        dlg.destroy()
        if response == gtk.RESPONSE_ACCEPT:
            return self.file_i_bins, self.file_i_bars, self.file_right_edge, self.file_sqrt, self.file_norm
        return None, None, None, None, None
    def __show_all_exponentials_info(self, *args):
        global ALL_EXPONENTIALS_MAX_N, ALL_EXPONENTIALS_AREA_CUTOFF, ALL_EXPONENTIALS_NEAR_THRESHOLD, ALL_EXPONENTIALS_SIGNIFICANCE, ALL_EXPONENTIALS_FINAL_CUTOFF
        result = qubx.pyenvGTK.prompt_entries(
            [
                ('Based on:', None, None, None),
                ('Exponential Sum-Fitting of Dwell-Time Distributions without Specifying Starting Parameters', None, None, None),
                ('David Landowne, Bin Yuan, and Karl L. Magleby', None, None, None),
                ('Biophysical Journal Volume 104 June 2013 2383-2391', None, None, None),
                ('Max components:', ALL_EXPONENTIALS_MAX_N, acceptIntGreaterThan(1), str),
                ('Area cutoff:', ALL_EXPONENTIALS_AREA_CUTOFF, acceptFloatGreaterThan(0), '%.3g'),
                ('Nearness threshold:', ALL_EXPONENTIALS_NEAR_THRESHOLD, acceptFloatGreaterThan(0), '%.3g'),
                ('Significance:', ALL_EXPONENTIALS_SIGNIFICANCE, acceptFloatGreaterThan(0), '%.2f'),
                ('Final area cutoff ("step 11"--enter 0 to skip):', ALL_EXPONENTIALS_FINAL_CUTOFF, acceptFloatGreaterThan(0), '%.3g')
            ],
            'About "Fit all exponentials"...'
        )
        if not (result is None):
            ALL_EXPONENTIALS_MAX_N = result[4]
            ALL_EXPONENTIALS_AREA_CUTOFF = result[5]
            ALL_EXPONENTIALS_NEAR_THRESHOLD = result[6]
            ALL_EXPONENTIALS_SIGNIFICANCE = result[7]
            ALL_EXPONENTIALS_FINAL_CUTOFF = result[8]

def events_per_scls(segs, mcls, scls): # [(model_classes, durations)]
    Ns = len(scls)
    per_scls = [ ([], []) for x in scls ]
    for cl, du in segs:
        for c, d in izip(cl, du):
            mc, sc = mcls[c]
            if mc >= 0:
                per_scls[sc][0].append(mc)
                per_scls[sc][1].append(d)
    return [(numpy.array(cc, dtype='int32'), numpy.array(dd, dtype='float32')) for cc, dd in per_scls]


def DurHist_Bins(data, td_ms, t_min=None, t_max=None, nbin=NBIN_DUR):
    tmax = t_max
    if tmax == None:
        tmax = max([ max([dd for cc,dd in itertools.izip(classes, durations) if cc==c]) for classes, durations in data ])
    tmax = tmax * 1e3 + td_ms
    tmin = t_min
    if tmin == None:
        tmin = min([ min([dd for cc,dd in itertools.izip(classes, durations) if cc==c]) for classes, durations in data ])
    tmin = tmin * 1e3 + td_ms
    N = nbin
    bins = zeros(shape=(N,), dtype='float32')
    if tmax == tmin:
        return bins
    dx = log10(tmax/tmin)/(N-1)
    r = pow(10.0, dx)
    bins[0] = tmin
    for i in xrange(1,N):
        bins[i] = bins[i-1] * r
    for i in reversed(range(1,N)):
        if bins[i] == bins[i-1]:
            bins[i-1:N-1] = bins[i:N]
            N -= 1
    bins.resize(N)
    return bins

def DurHist(c, data, td_ms, sampling, bins):
    N = len(bins)
    bars = zeros(shape=(N,))
    tot_count = 0
    for classes, durations in data:
        for i in xrange(len(durations)):
            if classes[i] != c: continue
            dur = durations[i] * 1e3 + td_ms
            last = 0.0
            for b in xrange(N):
                here = bins[b]
                if last < dur <= here:
                    bars[b] += 1.0
                last = here
            tot_count += 1
    if tot_count:
        bars /= tot_count
    return bins, bars

def DurHist_lib(c, data, td_ms, sampling, bins):
    cc, dd = data[0]
    return bins, durhist_calhist(c, cc, dd*1e3, bins, td_ms)

def DurHist_Smooth_lib(c, data, td_ms, sampling, bins):
    cc, dd = data[0]
    return bins, durhist_calhist_smooth(c, cc, dd*1e3, bins, td_ms, sampling)

def PDF(Q, clazz, a, td, bins, wantTcrit=True):
    together = zeros(shape=bins.shape)
    components = []
    taus = []
    amps = []

    try:
        Q = QtoQe(Q, clazz, td*1e-3)
        A = clazz == a
        Z = clazz != a
        Qaa = Q[ix_(A, A)]
        Qaz = Q[ix_(A, Z)]
        Qza = Q[ix_(Z, A)]
        Qzz = Q[ix_(Z, Z)]
        W, V = linalg.eig( (Qaa.I*Qaz*Qzz.I*Qza).T )
        i_unit = argmin(abs(W-1))
        pe = V[:,i_unit].T
        pe /= sum(pe)
    
        W, V = linalg.eig(Qaa)
        Vi = V.I
        spectre = matrix(zeros(shape=Qaa.shape))
        for i in xrange(Qaa.shape[0]):
            for j in xrange(Qaa.shape[0]):
                for k in xrange(Qaa.shape[0]):
                    spectre[j,k] = V[j,i] * Vi[i,k]
            coeff = sum(pe * spectre * Qaz) / W[i]
            itau = W[i]
            component = zeros(shape=bins.shape)
            last = bins[0] * bins[0] / bins[1]
            for b in xrange(len(bins)):
                next = bins[b]
                component[b] = coeff * (exp(itau*(next-td)*1e-3) - exp(itau*(last-td)*1e-3))
                together[b] += component[b]
                last = next
            components.append(component)
            taus.append(-1.0/itau)
            amps.append(-coeff)
    except linalg.LinAlgError:
        pass
    sortix = [i for i in xrange(len(taus))]
    sortix.sort(key=lambda i: taus[i])
    taus = [taus[i] for i in sortix]
    amps = [amps[i] for i in sortix]
    crits = SolveTCrits(taus, amps) if wantTcrit else []
    return together, components, taus, amps, crits

def SolveTCrits(taus, amps):
    #curve = qubx.fit.Curve("a1 * exp(-Tc / t1) - a2 * (1 - exp(-Tc / t2))")
    #curve = qubx.fit.Curve("a1 * (exp(-Tc / t1) - 1) + a2 * (exp(-Tc / t2) - 1) - a1*t1")
    curve = qubx.fit.Curve("a1 * exp(-Tc / t1) + a2 * exp(-Tc / t2) - a2")
    iA1, iT1, iA2, iT2, iTc = [curve.params.index(x) for x in ["a1", "t1", "a2", "t2", "Tc"]]
    for i in xrange(len(curve.params)):
        if i != iTc:
            curve.can_fit[i] = False
    fitter = qubx.fit.SimplexFitter()
    xx = yy = array([0.0], dtype='float32')
    crits = []
    params = [0.0] * len(curve.params)
    for i in xrange(len(taus)-1):
        params[iA1] = amps[i]
        params[iT1] = taus[i]
        params[iA2] = amps[i+1]
        params[iT2] = taus[i+1]
        params[iTc] = (taus[i] + taus[i+1]) / 2
        #curve.lo[iTc] = min(taus[i], taus[i+1])
        #curve.hi[iTc] = max(taus[i], taus[i+1])
        result, ssr, iterations, fit = fitter(curve, params, xx, yy)
        crits.append(result[iTc])
    return crits



intersect = lambda a0, a1, b0, b1: (max(a0, b0), min(a1, b1))
tri_rising_y = lambda peak, x: (x - peak + 1)
tri_falling_y = lambda peak, x: - ( x - peak - 1)
def tri_rising_area(peak, fro, to):
    x0, x1 = intersect(peak-1, peak, fro, to)
    if x0 < x1:
        y0 = tri_rising_y(peak, x0)
        y1 = tri_rising_y(peak, x1)
        return (0.5*(x1-x0)*(y0+y1))
    return 0.0
def tri_falling_area(peak, fro, to):
    x0, x1 = intersect(peak, peak+1, fro, to)
    if x0 < x1:
        y0 = tri_falling_y(peak, x0)
        y1 = tri_falling_y(peak, x1)
        return (0.5*(x1-x0)*(y0+y1))
    return 0.0
tri_area = lambda peak, fro, to: tri_rising_area(peak, fro, to) + tri_falling_area(peak, fro, to)

WAY_TOO_MANY_SAMPLES = 65536

def spread_events(bins, bars, samples, count, sampling_ms):
    if (count == 0) or (len(bins) == 0): return
    a = 0.0
    for i, b in enumerate(bins):
        bars[i] += count * tri_area(samples, a/sampling_ms, b/sampling_ms)
        a = b

def DurHist_Smooth(c, data, td_ms, sampling, bins):
    td = td_ms * 1e-3
    sampling_ms = sampling * 1e3
    N = len(bins)
    bars = zeros(shape=(N,))
    # first bin by how many samples,
    sampleCounts = []
    # but treat extra-long events separately so we don't need 200 million empty bins
    tot_count = 0
    for classes, durations in data:
        for i in xrange(len(durations)):
            if classes[i] != c: continue
            tot_count += 1
            samples = int(round((durations[i] + td) / sampling))
            if samples > WAY_TOO_MANY_SAMPLES:
                spread_events(bins, bars, samples, 1, sampling_ms)
            else:
                while len(sampleCounts) <= samples:
                    sampleCounts.append(0)
                sampleCounts[samples] += 1
    # then distribute events among the log bins
    # using empirical observation that events recorded as k samples long are actually between (k-1)dt and (k+1)dt,
    # distributed with mean of k*dt and linear falloff to 0 at +/- dt
    for samples, count in enumerate(sampleCounts):
        spread_events(bins, bars, samples, count, sampling_ms)
    if tot_count:
        bars /= tot_count
    return bins, bars



class DurHistCopyDialog(qubx.GTK.CopyDialog):
    def __init__(self, parent=None):
        qubx.GTK.CopyDialog.__init__(self, 'Copy histogram image', parent)
        line = gtk.HBox(True)
        line.show()
        lbl = gtk.Label('Class:')
        lbl.show()
        line.pack_start(lbl)
        self.txtClass = qubx.GTK.NumEntry(1, acceptInt)
        self.txtClass.set_width_chars(11)
        self.txtClass.show()
        line.pack_start(self.txtClass)
        self.vbox.pack_start(line, False, True)
    def run(self, view, Nclass):
        self.txtClass.accept = acceptIntBetween(1, Nclass)
        self.txtClass.value = min(Nclass, self.txtClass.value)
        response = qubx.GTK.CopyDialog.run(self, view)
        self.copy_class = self.txtClass.value - 1
        return response

TheDurHistCopyDialog = DurHistCopyDialog()


class SingleSignals(Requestable):
    def __init__(self, datasrc, Data, stimulus):
        Requestable.__init__(self)
        self.__ref = Reffer()
        self.datasrc = datasrc
        self.Data = Data
        self.datasrc.OnChangeIdealization += self.__ref(self.__onChangeIdl)
        self.__data = None
        self.stimulus = stimulus
        self.stimulus.OnInsert += self.__ref(self.__onChangeStimulus)
        self.stimulus.OnChange += self.__ref(self.__onChangeStimulus)
        self.stimulus.OnRemoving += self.__ref(self.__onChangeStimulus)
        self.Data.OnSwitch += self.__ref(self.__onSwitchDataFile)
        self.__datafile = None
        self.__onSwitchDataFile(self.Data, self.Data.file)
        self.__onChangeIdl()
    def qubx_finalize(self):
        pass
    def __onChangeIdl(self, *args):
        self.invalidate()
    def __onChangeStimulus(self, *args):
        self.invalidate()
    def __onSwitchDataFile(self, Data, file):
        if file == self.__datafile: return
        if self.__datafile:
            self.__datafile.constants.OnSet -= self.__ref(self.__onSetConstant)
        self.__datafile = file
        if self.__datafile:
            self.__datafile.constants.OnSet += self.__ref(self.__onSetConstant)
    def __onSetConstant(self, r, field, val, prev, undoing):
        if self.__datafile.constants[r,'Name'] == LBL_TDEAD:
            self.invalidate()
    def get(self, receiver):
        data = qubx.pyenv.env.globals['QubX'].Data.view
        if not data:
            gobject.idle_add(receiver, None, None, None, None) # signals, names, sampling_ms, td_ms
            return

        for i in xrange(data.file.constants.size):
            if LBL_TDEAD == data.file.constants.get(i, 'Name'):
                td = data.file.constants.get(i, 'Value')
                break
        else:
            td = data.file.sampling * 1e3
            data.file.constants.append({'Name':LBL_TDEAD, 'Value':td})
            data.file.constants.select(data.file.constants.size-1, 'Value', sender=self)

        segsrc = self.datasrc.get_segmentation()
        sig_names = [data.file.signals.get(segsrc[0].signal, 'Name')] if segsrc else []
        signals = [ [] ]
        for seg in segsrc:
            ff, ll, cc, dd = self.datasrc.get_idealization(seg, mark_excluded=True, get_fragments=True, get_durations=True)
            aa = data.file.ideal[segsrc[0].signal].seg[seg.index].amp
            signals[0].append( (cc, dd, aa) )
        def receive_stim(stim_names, stimidl):
            sig_names.extend(stim_names)
            for segs_flcda in stimidl:
                signals.append( [(cc, dd, aa) for ff,ll,cc,dd,aa in segs_flcda] )
            gobject.idle_add(receiver, signals, sig_names, qubx.pyenv.env.globals['QubX'].Data.file.sampling * 1e3, td)
        self.stimulus.call_with_stimulus_idl(receive_stim, [(seg.f, seg.l) for seg in segsrc])


class SinglePlex(Requestable):
    def __init__(self, signals):
        Requestable.__init__(self)
        self.__ref = Reffer()
        self.signals = signals
        self.signals.OnInvalidate += self.__ref(self.__onInvalidate)
    def __onInvalidate(self, signals):
        self.invalidate()
    def get(self, receiver):
        def receive_signals(signals, names, sampling_ms, td_ms):
            segs, mcls, scls = ProcessSignals(td_ms, signals)
            gobject.idle_add(receiver, segs, mcls, scls, names, sampling_ms, td_ms)
        self.signals.request(receive_signals)






ALL_EXPONENTIALS_MAX_N = 20
ALL_EXPONENTIALS_AREA_CUTOFF = 1e-5
ALL_EXPONENTIALS_NEAR_THRESHOLD = 0.02
ALL_EXPONENTIALS_FIT_REPS = 5
ALL_EXPONENTIALS_FIT_ITERS = 30
ALL_EXPONENTIALS_SIGNIFICANCE = 0.05
ALL_EXPONENTIALS_FINAL_CUTOFF = 1e-5

def llr_of_significance(sig):
    return log((1.0-.0005)/sig)

def ll_of_ssr(ssr, n):
    s2 = ssr/n
    return -(n/2)*(log(2*pi)+log(s2)) - ssr/(2*s2)

class AllExponentials(qubx.fit.BaseCurve):
    """This class implements a method to find all significant exponential components of a duration histogram, published in

Exponential Sum-Fitting of Dwell-Time Distributions without Specifying Starting Parameters
David Landowne,* Bin Yuan, and Karl L. Magleby*
Biophysical Journal Volume 104 June 2013 2383-2391
"""
    def __init__(self, expr="0.0", locals={}):
        qubx.fit.BaseCurve.__init__(self)
        self.name = 'AllExponentials'
        self.expr = "see Landowne et al, 2013"
        self.params = ["N"]
        for i in xrange(ALL_EXPONENTIALS_MAX_N):
            self.params.extend(["t%i" % i, "a%i" % i])
        self.lo = [UNSET_VALUE]*len(self.params)
        for i in xrange(ALL_EXPONENTIALS_MAX_N):
            self.lo[2*i+1] = 1e-16
            self.lo[2*i+2] = 0.0
        self.hi = [UNSET_VALUE]*len(self.params)
        self.can_fit = [False] * len(self.params)
        self.param_defaults = [1.0] * len(self.params)
        self.report_buf = StringIO()
        self.llr = 0.0
    def report(self, *args):
        self.report_buf.write('%s\n' % ' '.join(str(x) for x in args) )
    def eval(self, param_vals, xx, vvv=[]):
        # area under amp * exp(t / -tau):
        # amp * tau * (exp(-10**x{i-1} / -tau) - exp(-10**(r*x{i}) / tau))
        # where r = exp( log(10) * min(delta between two log10 x bins) )
        r = qubx.fit.calc_r(xx)
        xp = numpy.power(10.0, xx)
        xp /= r
        yy_all = numpy.zeros(shape=xx.shape, dtype=xx.dtype)
        for i in xrange(1, 1+2*max(0, min(ALL_EXPONENTIALS_MAX_N, int(round(param_vals[0])))), 2):
            tau, area = param_vals[i:i+2]
            if tau > 0.0:
                amp = area / tau
                xc = xp / (- tau)
                yy = numpy.exp(xc)
                xc *= r
                yy -= numpy.exp(xc)
                yy *= amp * tau
                yy_all += yy
        return yy_all
    # 1. Log bin the data (done by DurHist panel) -- xx contains log10(left bound of each bin)
    # 2. Generate the time constants for all possible exponentials
    def get_initial_params(self, xx, x0=None, x1=None):
        """Adjusts can_fit, returns initial param values spaced evenly between min(xx) and max(xx) on log axis."""
        N = ALL_EXPONENTIALS_MAX_N
        area = 1.0 / N
        params = [area] * len(self.params)
        params[0] = float(N)
        if x0 is None:
            X0, X1 = xx[0], xx[-1]
        else:
            X0, X1 = log10(x0), log10(x1)
        t0 = X0
        m = (X1 - X0) / (N - 1)
        for i in xrange(N):
            params[2*i+1] = 10**(t0 + i*m)
            self.can_fit[2*i+1] = False
            self.can_fit[2*i+2] = True
        return params
    # 3. Fit all areas with fixed time constants (done using a Fitter)
    # 4. Delete the end exponentials with insignificant area, generate new evenly spaced taus, fit their areas
    def get_params_minus_ends(self, params):
        """Adjusts can_fit, returns updated param values spaced evenly between smallest and largest legit expo in input params."""
        N = ALL_EXPONENTIALS_MAX_N # int(round(params[0]))
        i0 = 0
        while (i0 < N) and (params[2*i0+2] < ALL_EXPONENTIALS_AREA_CUTOFF):
            self.report('4. Deleting initial insignificant component at',params[2*i0+1])
            self.report(params)
            i0 += 1
        i1 = N-1
        while (i0 < i1) and (params[2*i1+2] < ALL_EXPONENTIALS_AREA_CUTOFF):
            self.report('4. Deleting final insignificant component at',params[2*i1+1])
            self.report(params)
            i1 -= 1
        if (i0 != 0) or (i1 != (N-1)):
            return self.get_initial_params(None, params[2*i0+1], params[2*i1+1])
        return params
    # 5. Delete all exponentials with negligible area
    def get_params_minus_negligible(self, params, cutoff, stepno):
        """Adjusts can_fit, returns updated param values with the smallest expos removed."""
        N = int(round(params[0]))
        params_out = [0] * len(params)
        i_out = 0
        for i_in in xrange(N):
            if params[2*i_in+2] >= cutoff:
                params_out[2*i_out+1], params_out[2*i_out+2] = params[2*i_in+1], params[2*i_in+2]
                i_out += 1
            else:
                self.report('%d. Deleting insignificant component at'%stepno,params[2*i_in+1])
                self.report(params)
        params_out[0] = float(i_out)
        for i in xrange(i_out):
            self.can_fit[2*i+1] = True
            self.can_fit[2*i+2] = True
        for i in xrange(i_out, ALL_EXPONENTIALS_MAX_N):
            self.can_fit[2*i+1] = False
            self.can_fit[2*i+2] = False
        return params_out
    # 6. Fit remaining areas and taus (using a Fitter)
    # 7. Combine exponentials with very close time constants (in order of decreasing time const, re-fitting after each)
    def get_params_combined(self, params, fitter, xx, yy, vvv, ww, ssr):
        """Adjusts can_fit, returns updated, fitted param values with near-identical time constants combined into one expo, and final ssr"""
        N = [int(round(params[0]))]
        params_out = params[:]
        # make sure they're in order of increasing tau
        expos = [(params[2*i+1], params[2*i+2]) for i in xrange(N[0])]
        expos.sort()
        for tau, area in expos:
            params_out[2*i+1], params_out[2*i+2] = tau, area

        def combine(i):
            t1, a1, t2, a2 = params_out[2*i+1:2*i+5]
            params_out[2*i+1] = t1*a1/(a1+a2) + t2*a2/(a1+a2)
            params_out[2*i+2] = a1+a2
            for j in xrange(i+1,N[0]-1):
                params_out[2*j+1:2*j+3] = params_out[2*j+3:2*j+5]
            N[0] -= 1
            params_out[0] = float(N[0])
            self.can_fit[2*N[0]+1] = self.can_fit[2*N[0]+2] = False
            params_out[2*N[0]+1] = params_out[2*N[0]+2] = 0.0
        for i in reversed(xrange(N[0]-1)):
            t0 = params_out[2*i+1]
            if abs((params_out[2*i+3] - t0) / t0) < ALL_EXPONENTIALS_NEAR_THRESHOLD:
                self.report('7. Combining close components at',t0,'and',params_out[2*i+3])
                self.report(params)
                combine(i)
                params_out[:], ssr, iterations, ff = fitter(self, params_out, xx, yy, vvv, ww)
        return params_out, ssr
    # 8. Determine the best fit with one fewer exponential
    def get_fit_with_fewer(self, params, fitter, xx, yy, vvv, ww, map_func):
        """One at a time, removes each exponential and re-fits; adjusts can_fit, returns reduced params, ssr of trial with max likelihood."""
        N = int(round(params[0]))
        ssr_params = map_func(all_expo_fit_without, [(xx, yy, vvv, ww, params, self.can_fit, i) for i in xrange(N)])
        ssr_without = [ssr for ssr, params_w in ssr_params]
        best_i = numpy.argmin(ssr_without)
        self.report('8. Removing component at',params[2*best_i+1])
        self.report(params)
        params_out = ssr_params[best_i][1]
        # then we consolidate the output parameters
        N -= 1
        for j in xrange(best_i, N):
            params_out[2*j+1:2*j+3] = params_out[2*j+3:2*j+5]
        params_out[0] = float(N)
        params_out[2*N+1] = params_out[2*N+2] = 0.0
        self.can_fit[2*N+1] = self.can_fit[2*N+2] = False
        return params_out, ssr_without[best_i]
    # 9. Repeat step 8 until it makes the likelihood of the fit significantly worse
    def try_fit_with_fewer(self, params, fitter, xx, yy, vvv, ww, ll0, map_func):
        params_fewer, ssr_fewer = self.get_fit_with_fewer(params, fitter, xx, yy, vvv, ww, map_func)
        ll = ll_of_ssr(ssr_fewer, len(xx))
        llr = ll0 - ll
        return llr, params_fewer, ssr_fewer
    # the following method was incorporated into run_generator() for responsiveness (more yielding)
    def get_minimal_fit(self, params, fitter, xx, yy, vvv, ww, ssr0, significance):
        """Adjusts can_fit, returns fully reduced params, ssr."""
        target_llr = llr_of_significance(significance or ALL_EXPONENTIALS_SIGNIFICANCE)
        ll0 = ll_of_ssr(ssr0, len(xx))
        if params[0] > 1:
            params_fewer, ssr = self.get_fit_with_fewer(params, fitter, xx, yy, vvv, ww)
            ll = ll_of_ssr(ssr, len(xx))
            llr = ll0 - ll
            self.llr = llr
            if llr <= target_llr:
                return self.get_minimal_fit(params_fewer, fitter, xx, yy, vvv, ww, ssr0, significance)
        N = int(round(params[0]))
        self.can_fit[2*(N-1)+1] = self.can_fit[2*(N-1)+2] = True # re-enable params from failed get_fit_with_fewer
        return params, ssr0
    # 10. Exclude brief erroneous exponentials -- with tau less than the dead time,
    #     and whose fitted area contributes < 1e-5 of the total fitted area
        # area * (exp(-10**xx[0] / -tau) - exp(-10**xx[-1] / tau))
    def get_params_minus_brief_errs(self, params, xx):
        """Returns params, stripped of brief erroneous components."""
        N = int(round(params[0]))
        params_out = params[:]
        left, right = 10**xx[0], 10**xx[-1]
        def fitted_area(i):
            tau, area = params[2*i+1], params[2*i+2]
            return area * (exp(-left / tau) - exp(-right / tau))
        i_out = 0
        for i in xrange(N):
            if (params[2*i+1] < left) and (fitted_area(i) < ALL_EXPONENTIALS_AREA_CUTOFF):
                self.report('10. Deleting brief erroneous component at',params[2*i+1])
                self.report(params)
                N -= 1
                self.can_fit[2*N+1] = self.can_fit[2*N+2] = False
                params_out[2*N+1] = params_out[2*N+2] = 0.0
            else:
                params_out[2*i_out+1:2*i_out+3] = params[2*i+1:2*i+3]
                i_out += 1
        params_out[0] = float(i_out)
        return params_out
    def run_generator(self, xx, yy, params, ww=None, fitter=None, map_func=map, significance=None):
        """Modifies params; yields percent complete after each step."""
        self.report_buf = StringIO()
        self.llr = 0.0
        f = qubx.fit.LMFitter() if (fitter is None) else fitter
        def fit(*args):
            for i in xrange(ALL_EXPONENTIALS_FIT_REPS):
                params, ssr, iterations, ff = f(*args)
                if iterations < ALL_EXPONENTIALS_FIT_ITERS:
                    break
            return params, ssr, iterations, ff
        vvv = []
        params[:] = self.get_initial_params(xx)
        yield 20.0
        DurHist = qubx.global_namespace.QubX.Figures.DurHist
        if DurHist.cancel_all_exp: return
        params[:], ssr, iterations, ff = fit(self, params, xx, yy, vvv, ww)
        yield 30.0
        if DurHist.cancel_all_exp: return
        params[:] = self.get_params_minus_ends(params)
        params[:], ssr, iterations, ff = fit(self, params, xx, yy, vvv, ww)
        yield 40.0
        if DurHist.cancel_all_exp: return
        params[:] = self.get_params_minus_negligible(params, ALL_EXPONENTIALS_AREA_CUTOFF, 5)
        yield 50.0
        if DurHist.cancel_all_exp: return
        params[:], ssr, iterations, ff = fit(self, params, xx, yy, vvv, ww)
        yield 60.0
        if DurHist.cancel_all_exp: return
        params[:], ssr = self.get_params_combined(params, fit, xx, yy, vvv, ww, ssr)
        yield 70.0
        if DurHist.cancel_all_exp: return

    # 9. Repeat step 8 until it makes the likelihood of the fit significantly worse
        target_llr = llr_of_significance(significance or ALL_EXPONENTIALS_SIGNIFICANCE)
        ll0 = ll_of_ssr(ssr, len(xx))
        N = int(round(params[0]))
        while N > 1:
            llr, params_fewer, ssr_fewer = self.try_fit_with_fewer(params, fit, xx, yy, vvv, ww, ll0, map_func)
            if llr > target_llr:
                self.report('...failed, llr =',llr,', keeping component')
                break
            self.report('...success, llr =',llr)
            self.llr = llr
            params[:], ssr = params_fewer, ssr_fewer
            N = int(round(params[0]))
            yield 90 - N
            if DurHist.cancel_all_exp: return
        self.can_fit[2*(N-1)+1] = self.can_fit[2*(N-1)+2] = True # re-enable params from failed get_fit_with_fewer

        yield 90.0
        if DurHist.cancel_all_exp: return
        params[:] = self.get_params_minus_brief_errs(params, xx)
        params[:], ssr, iterations, ff = fit(self, params, xx, yy, vvv, ww)
        yield 99.0
        if ALL_EXPONENTIALS_FINAL_CUTOFF: # "step 11"
            params[:] = self.get_params_minus_negligible(params, ALL_EXPONENTIALS_FINAL_CUTOFF, 11)
            params[:], ssr, iterations, ff = fit(self, params, xx, yy, vvv, ww)
        if DurHist.cancel_all_exp: return
        yield 100.0
    def run(self, xx, yy, ww=None, fitter=None, map_func=map, significance=None):
        """Returns fitted list of (tau, area) of all significant exponentials."""
        params = [1.0] * len(self.params)
        for pct in self.run_generator(xx, yy, params, ww, fitter, map_func, significance):
            pass
        result = []
        for i in xrange(int(round(params[0]))):
            result.append(tuple(params[2*i+1:2*i+3]))
        return result

class AllExponentials_Display(qubx.fit.BaseCurve):
    def __init__(self, expr='1 (sqrt) Exponential, to use: click "Fit all exponentials" in the DurHist panel', locals={}):
        qubx.fit.BaseCurve.__init__(self)
        self.name = 'AllExponentials'
        self.params = []
        self.lo = []
        self.hi = []
        self.can_fit = []
        self.param_defaults = []
        self.can_resample = False
        self.last_fit = []

        self.expr = expr # actually smuggling in N and sqrt
        self.N = int(re.search(r"([0-9]+) ", expr).group(1))
        self.sqrt = 'sqrt' in expr
        Ninv = 1.0 / self.N
        for i in xrange(self.N):
            self.params.extend(["tau%i"%(i+1), "area%i"%(i+1)])
            self.lo.extend([1e-16, 0.0])
            self.hi.extend([UNSET_VALUE, UNSET_VALUE])
            self.can_fit.extend([True, True])
            self.param_defaults.extend([(i+1)*Ninv, Ninv])
    def eval(self, param_vals, xx, vvv=[]):
        # area under amp * exp(t / -tau):
        # amp * tau * (exp(-10**x / -tau) - exp(-10**(r*x) / tau))
        # where r = exp( log(10) * min(delta between two log10 x bins) )
        r = qubx.fit.calc_r(xx)
        xp = numpy.power(10.0, xx)
        xp /= r
        yy_all = numpy.zeros(shape=xx.shape, dtype=xx.dtype)
        for i in xrange(self.N):
            tau, area = param_vals[2*i:2*i+2]
            if tau > 0.0:
                amp = area / tau
                xc = xp / (- tau)
                yy = numpy.exp(xc)
                xc *= r
                yy -= numpy.exp(xc)
                yy *= amp * tau
                yy_all += yy
        if self.sqrt:
            self.last_fit = numpy.sqrt(yy_all)
        else:
            self.last_fit = yy_all
        return self.last_fit
qubx.fit.Curves['AllExponentials'] = AllExponentials_Display

def all_expo_fit_without(args):
    try:
        # for more efficient test, we set area to 0.0 and can_fit to False
        xx, yy, vvv, ww, params, can_fit, i = args
        fitter = qubx.fit.LMFitter() # not generally possible to serialize a Fitter with multiprocessing, so ignoring caller's choice of fitter here
        p = params[:]
        p[2*i+2] = 0.0
        c = AllExponentials()
        c.can_fit[:] = can_fit[:]
        c.can_fit[2*i+1] = c.can_fit[2*i+2] = False
        fit_params, ssr, iterations, ff = fitter(c, p, xx, yy, vvv, ww)
        return ssr, fit_params
    except:
        traceback.print_exc()
        return 0.0
